/***********************************************************************/
/* Projekt : CAN-Intercepter (CAN-Logger)                              */
/* CPU : Renesas R8C/23, 20MHz                                         */
/*                                                                     */
/* Hauptprogramm                                                       */
/*                                                                     */
/*---------------------------------------------------------------------*/
/*                                                                     */
/* Copyright (c) 2007 Martin.S - martinsuniverse.de                    */
/*                                                                     */
/* Dieses Programm oder dieser Teil eines Programms darf verndert     */
/* und weitergegeben werden, solange dies nicht kommerziell geschieht. */
/* Die Nutzung ist nur zum Zweck der Fortbildung und zum  persnlichen */
/* Gebrauch gestattet. Weiterhin muss bei der Weitergabe und           */
/* Verwendung dieser Software der Copyright-Hinweis und die Lizenz-    */
/* bestimmung auch weitergegeben werden und gut erreichbar/abrufbar    */
/* sein.                                                               */
/*                                                                     */
/* Eine Gewhr fr die Funktionstchtigkeit und ein Haftung fr        */
/* Schden, die durch die Benutzung dieser Software entstehen, kann    */
/* nicht von bernommen werden. Es knnen also keine Garantieansprche */
/* geltend gemacht werden.                                             */
/*                                                                     */
/* Der Autor dieses Programms haftet nicht fr Schden an Soft- oder   */
/* Hardware oder Vermgensschden, die durch das Benutzen des          */
/* Programms entstehen, es sei denn diese beruhen auf einem grob       */
/* fahrlssigen oder vorstzlichen Handeln des Autors, seiner          */
/* Erfllungsgehilfen oder seiner gesetzlichen Vertreter.              */
/*                                                                     */
/*---------------------------------------------------------------------*/
/*                                                                     */
/* Historie:                                                           */
/*   10.10.2007  V0.1b  Erste lauffhige Beta                          */
/*                                                                     */
/***********************************************************************/
                  
#include "sfr_r823.h"
#include "msgsys.h"

#define STARTTAG 0xaa
#define ENDTAG 0xa5
#define RESETCMD 0x20
#define MASK 0x21
#define MASK14 0x22
#define MASK15 0x23
#define SLOTDISABLE 0xc0
#define SLOTTX 0x80
#define SLOTRX 0x40
#define CANERR 0x2f

#ifdef _KD_
	#define USBTXBUFFERSIZE 256				// nur 2^x Werte sind erlaubt
	#define USBRXBUFFERSIZE 32				// nur 2^x Werte sind erlaubt
#else
	#define USBTXBUFFERSIZE 1024			// nur 2^x Werte sind erlaubt
	#define USBRXBUFFERSIZE 64				// nur 2^x Werte sind erlaubt
#endif

#define USBTXBUFFERMASK USBTXBUFFERSIZE - 1
#define USBRXBUFFERMASK USBRXBUFFERSIZE - 1

typedef	union
{
		unsigned long l;
		unsigned int w[2];
		char b[4];
} uID;

typedef struct
{
	uID id;
	unsigned long timestamp;
	char dlc;
	char data[8];
} sRXCANRECORD;

typedef struct
{
	char cmd;
	uID id;
	char dlc;
	char data[8];
} sTXCANRECORD;

struct
{
	char tx_en:1;
	char rx_en:1;
} usbflags;

// Schnittstellenpuffer Variablen
char usbtxbuffer[USBTXBUFFERSIZE];					// Puffer fr Daten zum PC senden
sTXCANRECORD usbrxbuffer[USBRXBUFFERSIZE];	// Puffer fr Daten vom PC

unsigned int usbtxtop;		// ltester eintrag, muss zu erst verarbeitet werden
unsigned int usbtxbot;		// Einfgeposition
unsigned int usbrxtop;		// ltester eintrag, muss zu erst verarbeitet werden
unsigned int usbrxbot;		// Einfgeposition

// Variablen fr Dekodierung der Datenpackete vom PC
char msgtagfound;
char msgstarted;
char msgidlength;
char msgbytenum;
char msglength;

// Zeitgeber-Funktionen
unsigned int timestamp;

void timerinit(void);
void timerb_isr(void);

// Funktionen fr Schnittstelle zum PC
int usbinit(void);
int sendstring(const char *string);
int sendcmd(char *data, unsigned int size);
void usbrx_isr(void);
void usbtx_isr(void);

// Funtionen fr CAN-Controller
int caninit(char *cfg);
void encodeid(uID data, void *dest, char cmd);
void decodeid(void *src, uID *data, char cmd);
void canrx_isr(void);
void cantx_isr(void);
void canerr_isr(void);

sMSG msg;

unsigned int rxcount;

/***********************************************************************/
/* Hauptprogramm *******************************************************/
void main(void)
{
	prc0 = 1;
	cm1 = 0x08;
	cm0 = 0x00;
	asm("nop");
	asm("nop");
	asm("nop");
	asm("nop");
	ocd2 = 0;
	prc0 = 0;
	pur0 = 0xff;
	pd2 = 0xff;
	p2 = 0xff;
#ifndef _KD_
	pd6 = 0x40;
#endif
	asm("fset i");		// Interrupt freigeben, I-Flag setzen

	clearmsgcue();		// Nachrichtensystem initialisieren
	usbinit();				// Schnittstelle fr PC-Kommunikation initialisieren
	timerinit();
	
	// Hauptschleife
	do
	{	
		p2 ^= 0x01;
		// Prfen auf Nachrichten in der Warteschlange
		if (!getmsg(&msg))
		{
			// Auswerten der Nachricht in der Warteschlange
			switch(msg.msg)
			{
				// Datenpacket vom PC empfangen
				case UARTRX:
						{
	           	char cmd;
							
							// Kommando des Datenpackets umkopieren
							cmd = usbrxbuffer[usbrxtop].cmd;
							
							// Datenpacket auswerten
							switch (cmd & 0xef)
           		{
	            	// Reset Kommando empfangen
								case RESETCMD:
            				{
	            				char tempbuffer[2];
            				
											caninit(&usbrxbuffer[usbrxtop].data[0]);

	            				// Besttigung zum PC senden
											tempbuffer[0] = 2;
            					tempbuffer[1] = RESETCMD;
            					sendcmd(tempbuffer, tempbuffer[0]);
											
											rxcount = 0;
            				}
                  	break;  

								// Masken Daten empfangen
								case MASK:
										encodeid(usbrxbuffer[usbrxtop].id, &c0gmr.wa, cmd);
                 		break;  

								case MASK14:
										encodeid(usbrxbuffer[usbrxtop].id, &c0lmar.wa, cmd);
                 		break;  

								case MASK15:
										encodeid(usbrxbuffer[usbrxtop].id, &c0lmbr.wa, cmd);
                 		break;

             		default:
										// Weitere Auswertung des Datenpackets
										switch (cmd & 0xc0)
										{
											// CAN Slot deaktivieren
											case SLOTDISABLE:
													c0mctl[cmd & 0x0f].byte = 0x00;
													// Interrupt fr diesen einen Slot deaktivieren
													c0icr &= ~(1 << (cmd & 0x0f));
													break;
													
											// CAN Slot fr Empfang vorbereiten
											case SLOTRX:
													encodeid(usbrxbuffer[usbrxtop].id, &c0slot[cmd & 0x0f].wa.sidl, cmd);
													c0mctl[cmd & 0x0f].byte = cmd & 0xe0;

													// Standard / Extendet ID
													if (cmd & 0x10)
														c0idr |= 1 << (cmd & 0x0f);
													else
														c0idr &= ~(1 << (cmd & 0x0f));
														
													// Remote-Nachrich initialisieren
													if (cmd & 0x20)
													{
														c0slot[cmd & 0x0f].wa.dlc = usbrxbuffer[usbrxtop].dlc;

														if (usbrxbuffer[usbrxtop].dlc > 0)
														{
															char i;

															i = 0;
															while (i < usbrxbuffer[usbrxtop].dlc)
															{
																c0slot[cmd & 0x0f].wa.data[i ^ 0x01] = usbrxbuffer[usbrxtop].data[i];
																i++;
															}

															c0mctl[cmd & 0x0f].receive.rsplock = 0;
														}
														else
														{
															c0mctl[cmd & 0x0f].receive.rsplock = 1;
														}
													}
														
													
													// Interrupt fr diesen einen Slot aktivieren
													c0icr |= 1 << (cmd & 0x0f);
													break;
													
											// CAN Nachricht senden
											case SLOTTX:
													// Prfen, ob Slot gerade in Benutzung
													if ((c0mctl[cmd & 0x0f].byte & 0x82) == 0x82)
													  while (c0mctl[cmd & 0x0f].transmit.trmactive);  // Warten wenn Slot nicht nutzbar
													
													// Control register initialisieren
													while (c0mctl[cmd & 0x0f].byte)
													{
														c0mctl[cmd & 0x0f].byte = 0x00;
													}
													
													// Nachricht in den Slot kopieren
													encodeid(usbrxbuffer[usbrxtop].id, &c0slot[cmd & 0x0f].wa.sidl, cmd);
													c0slot[cmd & 0x0f].wa.dlc = usbrxbuffer[usbrxtop].dlc;
													{
														char i;
														
														i = 0;
														while (i < usbrxbuffer[usbrxtop].dlc)
														{
															c0slot[cmd & 0x0f].wa.data[i ^ 0x01] = usbrxbuffer[usbrxtop].data[i];
															i++;
														}
													}
													
													// Standard / Extendet ID
													if (cmd & 0x10)
														c0idr |= 1 << (cmd & 0x0f);
													else
														c0idr &= ~(1 << (cmd & 0x0f));
													
													if (cmd & 0x20)
														// Interrupt fr diesen einen Slot aktivieren
														c0icr |= 1 << (cmd & 0x0f);
													else
														// Interrupt fr diesen einen Slot deaktivieren
														c0icr &= ~(1 << (cmd & 0x0f));
													
  												// Senden starten
													c0mctl[cmd & 0x0f].byte = cmd & 0x20;
													c0mctl[cmd & 0x0f].transmit.trmreq = 1;
													break;
             				}	// switch (cmd & 0xc0)
           		}	// switch (cmd)
							asm("fclr i");			// Interrupt sperren
							usbrxtop++;
							usbrxtop &= USBRXBUFFERMASK;
							asm("fset i");			// Interrupt sperren

							break;
						}	// case UARTRX:
			}	// switch(msg.msg)
		}
	}		// Hauptschleife
	while(1);
}
/* Hauptprogramm Ende **************************************************/
/***********************************************************************/


//*****************************************************************************
// Funtion : 
// Gltigkeit : global
// Parameter : -
// Rckgabe : -
// ToDo : -
//*****************************************************************************
void timerinit(void)
{
	// Zeitgeber stoppen
	tstart_tracr = 0;		// Timer A stoppen
	tstart_trbcr = 0;		// Timer B stoppen

	timestamp = 0;

	// Zeitgeber A initialisieren
	tramr = 0x10;				// Timer mode, f8 source
	traioc = 0x00;			// IO Control
	trapre = 24;				// Vorteiler (+1)
	tra = 99;						// Reload (+1)
	
	// Zeitgeber B initialisieren
	trbmr = 0x20;				// Timer mode, Timer A source
	trbioc = 0x00;			// IO Control
	trbpre = 255;				// Vorteiler (+1)
	trbpr = 255;				// Reload (+1)
	trbic = 1;					// Interrupt Prioritt
	
	// Zeitgeber starten
	tstart_tracr = 1;		// Timer A starten
	tstart_trbcr = 1;		// Timer B starten
}

//*****************************************************************************
// Funtion : 
// Gltigkeit : global
// Parameter : -
// Rckgabe : -
// ToDo : -
//*****************************************************************************
#pragma INTERRUPT timerb_isr
void timerb_isr(void)
{
	p2 ^= 0x02;
	timestamp++;
}

//*****************************************************************************
// Funtion : Asyncrone Schnittstelle fr USB initialisieren
// Gltigkeit : global
// Parameter : -
// Rckgabe : Fehlercode
// ToDo : -
//*****************************************************************************
int usbinit(void)
{
	usbrxbot = usbrxtop = usbtxbot = usbtxtop = 0;
	
	usbflags.tx_en = 1;
	msgstarted = 0;
	msgtagfound = 0;
	
#ifdef _KD_
	// UART0
//	u0brg = 0x00;		// 1,25Mbps
	u0brg = 0x01;		// 625kbps
	u0mr = 0x05;		// 8n1
	u0c0 = 0x00;
	u0c1 = 0x05;		// enable Tx und Rx
	
	// interrupt freigabe
	s0tic = 0x05;
	s0ric = 0x05;
#else
	// UART1 - USB-Dongle
	pmr = 0x10;			// Port 6.6 u. 6.7 UART1 Funktion
	u1sr = 0x03;
//	u1brg = 0x00;		// 1,25Mbps
	u1brg = 0x01;		// 625kbps
	u1mr = 0x05;		// 8n1
	u1c0 = 0x00;
	u1c1 = 0x05;		// enable Tx und Rx
	
	// interrupt freigabe
	s1tic = 0x05;
	s1ric = 0x05;
#endif
}

//*****************************************************************************
// Funtion : Asyncrone Schnittstelle fr USB initialisieren
// Gltigkeit : global
// Parameter : -
// Rckgabe : Fehlercode
// ToDo : -
//*****************************************************************************
#pragma INTERRUPT usbrx_isr
void usbrx_isr(void)
{
	char rxchar;
	sMSG msg;

#ifdef _KD_
	rxchar = (unsigned int)u0rb;
#else
	rxchar = (unsigned int)u1rb;
#endif

	p2 ^= 0x04;

  if ((rxchar == STARTTAG) || (rxchar == ENDTAG))
  {
    if (msgtagfound != rxchar)
    {
      msgtagfound = rxchar;
      if (rxchar == ENDTAG)
      {
        // Prfen, ob Nachrichtenlnge past
        if ((msgbytenum == msglength) && msgstarted)
        {
          // Daten Nachricht speichern
					usbrxbot++;
					usbrxbot &= USBRXBUFFERMASK;

					msg.msg = UARTRX;
					addmsg(&msg);

          msgstarted = 0;
          msgtagfound = 0;

					p2 ^= 0x08;
        }
      }
    }
    else
    {
      msgtagfound = 0;
    }
  }
  else
  {
    if (msgtagfound == STARTTAG)
    {
      msgstarted = 1;
      msgbytenum = 0;
      usbrxbuffer[usbrxbot].id.l = 0;
      usbrxbuffer[usbrxbot].dlc = 0;
    }
    if (msgtagfound == ENDTAG)
      msgstarted = 0;

    msgtagfound = 0;
  }

  if (msgbytenum > msglength)
    msgstarted = 0;

  if (!msgtagfound && msgstarted)
  {
    if (!msgbytenum)
     	msglength = rxchar;
    
    if (msgbytenum == 1)
    {
			usbrxbuffer[usbrxbot].cmd = rxchar;
			// Prfen auf Reset/Ini Kommando
			if (rxchar == 0x20)
 				msgidlength = 0;
			else
				// Lnge der ID prfen
				if (rxchar & 0x10)
   				msgidlength = 4;
				else
     			msgidlength = 2;
    }
     	
    if (msgbytenum > 1)
		{
			if (msgbytenum < (msgidlength + 2))
				// ID herausfiltern
	    	usbrxbuffer[usbrxbot].id.b[msgbytenum - 2] = rxchar;
			else
			{
	    	// Datenbytes
	 			usbrxbuffer[usbrxbot].data[usbrxbuffer[usbrxbot].dlc] = rxchar;
 				usbrxbuffer[usbrxbot].dlc++;
			}
		}
	  msgbytenum++;
  }
}

//*****************************************************************************
// Funtion : Asyncrone Schnittstelle fr USB initialisieren
// Gltigkeit : global
// Parameter : -
// Rckgabe : Fehlercode
// ToDo : -
//*****************************************************************************
#pragma INTERRUPT usbtx_isr
void usbtx_isr(void)
{
	if ((usbtxtop != usbtxbot) && (usbflags.tx_en))
	{
#ifdef _KD_
		u0tbl = usbtxbuffer[usbtxtop++];
#else
		u1tbl = usbtxbuffer[usbtxtop++];
#endif
		usbtxtop &= USBTXBUFFERMASK;
	}
}

//*****************************************************************************
// Funtion : Asyncrone Schnittstelle fr USB initialisieren
// Gltigkeit : global
// Parameter : -
// Rckgabe : Fehlercode
// ToDo : -
//*****************************************************************************
int sendstring(const char *string)
{
	while(*string)
	{
		usbtxbuffer[usbtxbot++] = *string++;
		usbtxbot &= USBTXBUFFERMASK;
	}

#ifdef _KD_
	if (ti_u0c1)
	{
		u0tbl = usbtxbuffer[usbtxtop++];
#else
	if (ti_u1c1)
	{
		u1tbl = usbtxbuffer[usbtxtop++];
#endif
		usbtxtop &= USBTXBUFFERMASK;
	}
	
	return OK;
}

//*****************************************************************************
// Funtion : Asyncrone Schnittstelle fr USB initialisieren
// Gltigkeit : global
// Parameter : -
// Rckgabe : Fehlercode
// ToDo : -
//*****************************************************************************
int sendcmd(char *data, unsigned int size)
{
  char i;

	usbflags.tx_en = 0;
	usbtxbuffer[usbtxbot++] = STARTTAG;
	usbtxbot &= USBTXBUFFERMASK;
  i = 0;
  while (i++ < size)
  {
		usbtxbuffer[usbtxbot++] = *data;
		usbtxbot &= USBTXBUFFERMASK;
    if ((*data == 0xaa) || (*data == 0xa5))
		{
			usbtxbuffer[usbtxbot++] = *data;
			usbtxbot &= USBTXBUFFERMASK;
		}
		data++;
  }
	usbtxbuffer[usbtxbot++] = ENDTAG;
	usbtxbot &= USBTXBUFFERMASK;

	usbflags.tx_en = 1;
#ifdef _KD_
	if (ti_u0c1)
	{
		u0tbl = usbtxbuffer[usbtxtop++];
#else
	if (ti_u1c1)
	{
		u1tbl = usbtxbuffer[usbtxtop++];
#endif
		usbtxtop &= USBTXBUFFERMASK;
	}
}

//*****************************************************************************
// Funtion : CAN-Controller 0 initialisieren
// Gltigkeit : global
// Parameter : -
// Rckgabe : Fehlercode
// ToDo : -
//*****************************************************************************
int caninit(char *cfg)
{
  reset_c0ctlr = 1;             // Reset/initialization mode
  sleep_c0ctlr = 0;
  while (!state_reset_c0str);   // Warten bis CAN-Controller im Init-Mode
  
	// Takterzeugung/Taktteilung
	prc0 = 1;                     // Schreibschutz aufheben
  cclkr = cfg[1];         			// fCAN = f/1
  prc0 = 0;                     // Schreibschutz setzen

  // Reset/Initialisation verlassen
  reset_c0ctlr = 0;             // Reset/initialization verlassen
  while (state_reset_c0str);    // Warten Modus wechsel

  c0ctlrl = (cfg[0] & 0x1a) | 0x40;	// Port-Pins fr CAN, Word access
  c0ctlrh = (cfg[0] & 0x20) | 0x03;

  // Bautrate setzen
  c0conrl = cfg[2];             // Baudrate, SJW, PTS, PBS1, PBS2
  c0conrh = cfg[3];             // Baudrate, SJW, PTS, PBS1, PBS2
	
  // CAN-Interrupt Steuerung initialisieren
  c0recic = 0x05;				        // Rx-Int-Level, CAN0 Message Rx Complete interrupt
//	c01erric = 0x05;
//  c0trmic = 0x05;				       	// Tx-Int-Level, CAN0 Message Tx Complete interrupt
}

//*****************************************************************************
// Funtion : Asyncrone Schnittstelle fr USB initialisieren
// Gltigkeit : global
// Parameter : -
// Rckgabe : Fehlercode
// ToDo : -
//*****************************************************************************
void encodeid(uID data, void *dest, char cmd)
{
	*(unsigned int*)dest = ((data.w[0] << 2) & 0x1f00) | (data.w[0] & 0x003f);
	if (cmd & 0x10)
	{
		((unsigned int*)dest)++;
		*(unsigned int*)dest = data.l >> 17;
		((unsigned int*)dest)++;
		*(unsigned int*)dest = (data.l >> 3) & 0x3f00;
	}
}

//*****************************************************************************
// Funtion : Asyncrone Schnittstelle fr USB initialisieren
// Gltigkeit : global
// Parameter : -
// Rckgabe : Fehlercode
// ToDo : -
//*****************************************************************************
void decodeid(void *src, uID *data, char cmd)
{
	data->l = (*(char*)src) & 0x3f;
	((char*)src)++;
	data->l |= ((*(char*)src) & 0x1f) << 6;
	if (cmd & 0x10)
	{
		((char*)src)++;
		data->l |= ((unsigned long)(*(unsigned int*)src) & 0x0fff) << 17;
		((unsigned int*)src)++;
		data->l |= ((unsigned long)(*(unsigned int*)src) & 0x3f00) << 3;
	}
}

//*****************************************************************************
// Funtion : 
// Gltigkeit : global
// Parameter : -
// Rckgabe : -
// ToDo : -
//*****************************************************************************
#pragma INTERRUPT canrx_isr
void canrx_isr(void)
{
  char slot;
	char i;
	char j;
	char tempbuffer[17];
	char tempmctl;

	slot = c0str & 0x0f;

	tempbuffer[0] = 7;						// Nachrichtenlnge
	tempbuffer[1] = SLOTRX | slot;
	if ((c0idr >> slot) & 0x0001)
	{
		tempbuffer[0] += 2;
		tempbuffer[1] |= 0x10;
	}
	tempbuffer[1] |= c0mctl[slot].byte & 0x20;
	tempbuffer[2] = ~trbpre;			// Zeitstempel
	tempbuffer[3] = ~trbpr;				// Zeitstempel
	tempbuffer[4] = timestamp ;		// Zeitstempel
	decodeid(&c0slot[slot].wa.sidl, (uID*)&tempbuffer[5], 0);

	i = 0;
	j = tempbuffer[0];
	while (i < c0slot[slot].wa.dlc)
	{
		tempbuffer[j++] = c0slot[slot].wa.data[i++ ^ 0x01];
	}
	tempbuffer[0] += c0slot[slot].wa.dlc;				// Nachrichtenlnge
	tempmctl = c0mctl[slot].byte;
	while (c0mctl[slot].byte)
	{
		c0mctl[slot].byte = 0x00;
	}
	c0mctl[slot].byte = tempmctl & 0x70;

	p2 ^= 0x10;
	sendcmd(tempbuffer, tempbuffer[0]);
}

//*****************************************************************************
// Funtion : 
// Gltigkeit : global
// Parameter : -
// Rckgabe : Fehlercode
// ToDo : -
//*****************************************************************************
#pragma INTERRUPT cantx_isr
void cantx_isr(void)
{
}

//*****************************************************************************
// Funtion : 
// Gltigkeit : global
// Parameter : -
// Rckgabe : Fehlercode
// ToDo : -
//*****************************************************************************
#pragma INTERRUPT canerr_isr
void canerr_isr(void)
{
	char tempbuffer[5];

	p2 ^= 0x20;

	tempbuffer[0] = 5;				// Nachrichtenlnge
	tempbuffer[1] = CANERR;
	tempbuffer[2] = c0strh;		// Status
	tempbuffer[3] = c0rec;		// Receice error count
	tempbuffer[4] = c0tec;		// Transmit error count

	sendcmd(tempbuffer, tempbuffer[0]);
}
